'' +--------------------------------------------------------------------------+
'' | Cluso's CPM2.2 to/from FAT16/32 File transfer                     v0.002 |
'' +--------------------------------------------------------------------------+
'' |  Author:            "Cluso99" (Ray Rodrick)                              |
'' |  Copyright (c) 2012 "Cluso99" (Ray Rodrick)                              |
'' |  License            MIT License - See end of file for terms of use       |
'' |  Requires:          SD-MMC_FATEngine.spin by Kye                         |
'' +--------------------------------------------------------------------------+
''                             
'' RR20120121   001a  Copy a file from CPM2.2 to FAT16/32 (text files for now)
'' RR20120124   001n  Reads CPM file ok (less notes below), only 1 dir entry!
'' RR20120125   002a  First sample release using fixed filenames
''                 b  allow multiple dir entries
'' RR20120125   003   Working with multiple dir entries (max 1MB files)
'' RR20120126   004   Don't worry about Ctl-Z so works with text and binaries


'  TO DO:
'  Does not copy readonly or hidden files.
'  Only copies files from user 0 (no check for the time being)
'??  Need to scan backwards for Ctl-Z and only last record, not last record in block too (bug)!
'  Does not request filenames/disknames or direction
'  Does not insert date/time stamp


CON

  _clkmode = xtal1 + pll16x     
  _xinfreq = 6_500_000          


'RamBlade3 hardware...
  SI            = 31
  SO            = 30
  baud          = 115_200

  _dopin        = 20
  _clkpin       = 21
  _dipin        = 22
  _cspin        = 23
  _cdpin        = -1 ' -1 if unused.
  _wppin        = -1 ' -1 if unused.

  _rtcres1      = -1 ' -1 always.
  _rtcres2      = -1 ' -1 always.  
  _rtcres3      = -1 ' -1 always.  



''      CPM2.2 File System
''      ------------------
''      'ZiCog 8MB Hard Disk Parameter Blocks: 17 bytes CP/M + 2 bytes sector size used by SIMH hdsk_dpb
''      dpb_0                           '8MByte Default SIMH Altair HDSK params
''      :spt_low      BYTE      $20     'sectors per track (low byte)                   \ 32x 128byte sectors
''      :spt_high     BYTE      $00     'sectors per track (high byte)                  /
''      :bsh          BYTE      $05     'data allocation Block SHift factor               5=4KB  
''      :blm          BYTE      $1F     'data allocation block mask                       $1F=4KB
''      :exm          BYTE      $01     'EXtent Mask
''      :dsm_low      BYTE      $F9     'maximum data block number (low_byte)           \ $07F9+1= 2048-6 (8MB/4KB=2K -off)
''      :dsm_high     BYTE      $07     'maximum data block number (high_byte)          /
''      :drm_low      BYTE      $FF     'total number of directory entries (low byte)   \ $03FF+1= 1024 entries
''      :drm_high     BYTE      $03     'total number of directory entries (high byte)  /          (32 bytes)
''      :al0          BYTE      $FF     'determine reserved directory blocks            \ $00FF+1= 256 blocks
''      :al1          BYTE      $00     'determine reserved directory blocks            /
''      :cks_low      BYTE      $00     'size of directory ChecK vector (low byte)      \ $0000 for fixed disk
''      :cks_high     BYTE      $00     'size of directory ChecK vector (high byte)     /
''      :off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)  \ $0006 
''      :off_high     BYTE      $00     'number of reserved tracks (offset) (high byte) /
''      :psh          BYTE      $00     'Physical record SHift factor, CP/M 3           \ $0000
''      :phm          BYTE      $00     'PHhysical record Mask, CP/M 3                  /
''      :ss_low       BYTE      $80     'Sector Size (low byte)                         \ $0080= 128 byte sectors
''      :ss_high      BYTE      $00     'Sector Size (high byte)                        /
''                                      'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.

'The folowing are for CPM2.2 & "the" 8MB Hard Disk as we have setup
  sectorspertrack = $0020       'spt = 32
  blockshift      = 5           'bsh = 4KB
  blockmask       = $1F         'blm = 4KB
  extentmask      = $01         'exm = 1
  maxblock        = $07F9       'dsm = 2048 -offsettracks -1   (8MB/4KB=2K -off -1)
  direntries      = $03FF+1     'drm +1 
  reservblocks    = $00FF+1     'al0:al1 +1
'                 = $0000       'cks
  offsettracks    = $0006       'off
'                 = $00         'psh
'                 = $00         'phm
  sectorsize      = $80         'ss = 128
    
  dirbase         = offsettracks * sectorspertrack '=192  'base of cpm directory entries (128 byte sectors)
  dirsize         = 32          'directory entry size = 32 bytes
  filebase        = direntries * dirsize/sectorsize + dirbase '=1024*32/128+192=448  'base of file data (end of dir entries)
  dirdeleted      = $E5         'directory entry unused/deleted
  recsperblock    = 1<< blockshift '1<<5=32

  MAXEXTENTS      = 32          'maximum no of directory entries (extents) for a file (=1MB)

''      
''      Directory Entries (32 bytes)
''      ----------------------------
''      UU  byte        'user number 0-15 or 0-31, $E5=unused/deleted
''      FN  byte[8]     'filename
''      TY  byte[3]     'filetype b7 of TY[0]=1=readonly, TY[1]=1=hidden
''      EX  byte        'extent counter (low) 0-31 (bits0-4) see S2 for (bits 5-12)
''                      ' where:   dir-entry-no = ((32*S2)|EX)/(exm+1) = ((S2<<5)|EX)<<1
''      S1  byte        'reserved =0
''      S2  byte        'extent counter (high) (bits 5-12)
''      RC  byte        'no of records used in this extent (128 bytes)
''                      'total records this extent
''                      'where:    recs = ((EX & exm)*128)+RC = ((EX & $01)*128)+RC
''      AL  word[8]     '8 allocation block locations (of 4KB for zicog 8MB HD)
''                      'where the block no is from the dir base (sector 192 for zicog 8MB HD)
''      'There is no allocation map (FAT) and is calculated at CPM boot from the dir entries.                             

{
DRAC_A  .DSK open 897 8388608
DIR at 24576
946 00...01 00 00 80 40 00 41 00 42 00 46 00 47 00 48 00 49 00 4A 00 .BDOS    PRN....@.A.B.F.G.H.I.J.
946 00...03 00 00 80 4B 00 4C 00 4D 00 4E 00 4F 00 50 00 51 00 52 00 .BDOS    PRN....K.L.M.N.O.P.Q.R.
946 00...05 00 00 80 53 00 54 00 55 00 56 00 57 00 58 00 59 00 5A 00 .BDOS    PRN....S.T.U.V.W.X.Y.Z.
947 00...07 00 00 80 5B 00 5C 00 5D 00 5E 00 5F 00 60 00 61 00 62 00 .BDOS    PRN....[.\.].^._.`.a.b.
947 00...09 00 00 80 63 00 64 00 65 00 66 00 67 00 68 00 69 00 80 00 .BDOS    PRN....c.d.e.f.g.h.i...
947 00...0A 00 00 40 81 00 82 00 00 00 00 00 00 00 00 00 00 00 00 00 .BDOS    PRN...@................
1008
}


OBJ

  fat0: "SD-MMC_FATEngine.spin"
  fat1: "SD-MMC_FATEngine.spin"
  fdx:  "FullDuplexSerial_rr004"
  
VAR
  byte  databuf[512]
  long  found, records, allocblock, findrecord, foundaddr, total_records
  long  uu,ex,s1,s2,rc,al[8]
  long  dir_position[MAXEXTENTS]                'store the position of the entry (seek posn)
  byte  dir_sequence[MAXEXTENTS]                'store the directory sequence no
  byte  dir_sort[MAXEXTENTS]                    'used to sort the dir entries into sequence
  long  dir_index                               'index to dir_position & dir_sequence 
  

DAT
' diskname0  byte  "ZICOG_A2.DSK",0             'cpm disk
  diskname0  byte  "DRAC_A  .DSK",0             'cpm disk
' filename0  byte  "MYNAME  BAS",0              'cpm
  filename0  byte  "BDOS    PRN",0              'cpm
'  filename0  byte  "XMODEM  TXT",0              'cpm
  filename1  byte  "FROMCPM .TXT",0              'prop
  
PUB main | errorString

  waitcnt(clkfreq*5 + cnt)        'wait 1 sec
  fdx.start(SI, SO, 0, baud)
  waitcnt(clkfreq/2 + cnt)
  fdx.tx(0)                                             'cls
  fdx.str(string("CPM XFER v003 ",13))
                          
  ' "fatEngineStart" must only be called once.
  fat0.fatEngineStart(_dopin,_clkpin,_dipin,_cspin,_wppin,_cdpin,_rtcres1,_rtcres2,_rtcres3)


  errorString := \readCPM ' Returns the address of the error string or null.

  if(errorString)
    fdx.str(string("Error: "))
    StopOnError(errorString)
    
  if(fat0.partitionMounted)
    \fat0.unmountPartition
  if(fat1.partitionMounted)
    \fat1.unmountPartition

  waitcnt(clkfreq + cnt)
  fat0.fatEngineStop            ' Give the block driver a second to finish up.

  fdx.str(string("      Records= "))
  fdx.dec(total_records)
  fdx.str(string(" ("))
  total_records >>= 3                                   ' /8 converts to KB
  fdx.dec(total_records)
  fdx.str(string("KB)"))
  crlf
  
  fdx.str(string("Copy complete."))
  crlf
  
  repeat                                                '<===== loop here

'PRI readCPM | errorString, errorNumber, buffer, counter, data[128], i, j, k, tmp
PRI readCPM | errorString, errorNumber, i, j, k, tmp

  fat0.mountPartition(0)
  fat1.mountPartition(0)

'-----------------------------------------------------------------
'Open input file (diskname0)    This is a cpm disk stored as a FAT file.
  fdx.str(@diskname0)
  fat0.openFile(@diskname0, "R")                         'cpm diskname 
  fdx.str(string(" open "))
  fdx.dec(fat0.getSector)
  fdx.tx(" ")
  fdx.dec(fat0.filesize)
  fdx.tx(13)

  fdx.str(string("dirbase="))
  fdx.dec(dirbase)
  fdx.str(string(" direntries="))
  fdx.dec(direntries)
  fdx.str(string(" filebase="))
  fdx.dec(filebase)
  fdx.str(string(" sectorsize="))
  fdx.dec(sectorsize)
  fdx.str(string(" dirsize= "))
  fdx.dec(dirsize)
  fdx.tx(13)

'-----------------------------------------------------------------
'Create a new output file (file1)
  errorString := \fat1.newFile(@filename1)
  errorNumber :=  fat1.partitionError                   ' Returns zero if no errors

  if(errorNumber)                                       ' Handle already exists error
    if(errorNumber == fat1#Entry_Already_Exist)
      ' Re-create the file.
      fat1.deleteEntry(@filename1)
      errorString := fat1.newFile(@filename1)
      errorNumber :=  fat1.partitionError               ' Returns zero if no errors
      if(errorNumber)
        abort errorString       ' Unable to handle the error. Continue passing the error.
    else
      abort errorString         ' Unable to handle the error. Continue passing the error.

  fat1.openFile(@filename1, "W")                        ' open for write

'-----------------------------------------------------------------
'SEEK TO THE FIRST DIR ENTRY
  i:=fat0.fileSeek(dirbase * sectorsize)                'seek to first DIR entry (byte) position
' fdx.str(string("DIR at "))
' fdx.dec(i)
' fdx.tx(13)

'NOW DISPLAY (ALL/SPECIFIC) FILENAME ENTRIES
  dir_index := 0                                        'reset dir_position/seqence index
  repeat i from 0 to direntries-1
    tmp := fat0.getCurrentPosition                      'save posn in case we found the entry
    fat0.readData(@databuf, dirsize)                    'read 32 byte dir entry
    if databuf[0] <> dirdeleted                         '$E5 = no entry so skip
      found:=true                                       'preset
      repeat j from 0 to 10                             'validate filename entry
        if databuf[j+1] <> filename0[j]
          found:=false                                  'not found
'     found:=true                                       'display all !!!
      if found                                          'display if entry found 
        'add the dir entry to the table
        dir_position[dir_index] := tmp                  'save the position
        dir_sequence[dir_index] := ((databuf[14]<<5) | databuf[12])/(extentmask +1) '((S2<<5)|EX)/(exm+1)
        dir_index++
        'display
        foundaddr := tmp                                'save the posn for later
        fdx.dec(dir_position[dir_index-1])
        fdx.str(string(" seq# "))
        fdx.dec(dir_sequence[dir_index-1])
'        fdx.dec(foundaddr)
        fdx.tx(" ")
        fdx.dec(fat0.getSector)
        fdx.tx(" ")
        fdx.hex(databuf[0],2)                         
        fdx.str(string("..."))
        repeat j from 12 to dirsize-1
          fdx.hex(databuf[j],2)                         
          fdx.tx(" ")
        repeat j from 0 to dirsize-1
          fdx.tx(ascii(databuf[j]))
        fdx.tx(13)
        if dir_index => MAXEXTENTS
          StopOnError(string("Too many extents."))
  fdx.dec(fat0.getSector)                               'verify we are at the end
  fdx.tx(13)
  fdx.tx(13)

'-----------------------------------------------------------------
'NOW VALIDATE THE DIR ENTRIES SEQUENCE NUMBERS.
  'now sort the entries by placing the entry# in the sort# as offset by the seq#
  repeat i from 0 to dir_index-1
    j := dir_sequence[i]                                'get the seq#
    if j => dir_index                                   'check it is not too big
      StopOnError(string("Extent sequences invalid."))
    dir_sort[j] := i                                    'store dir# in the sort# offset by seq#
  repeat i from 0 to dir_index-1
    fdx.dec(dir_sort[i])
    fdx.tx(" ")
  fdx.tx(13)  
'-----------------------------------------------------------------
'HERE IF SEQUENCE NUMBERS VALID. PROCESS EACH DIR ENTRY IN SEQUENCE (using sort#)      
  total_records := 0
  repeat i from 0 to dir_index-1
    process_one_entry(dir_position[dir_sort[i]])        'process this sector position 

'-----------------------------------------------------------------
  fat0.closeFile
  fat0.unmountPartition
  fat1.closeFile
  fat1.unmountPartition
'-----------------------------------------------------------------

PRI process_one_entry(seek_position) | i, j, tmp
'' Process the dir entry at seek_position

'-----------------------------------------------------------------
'SEEK TO THE DIR ENTRY
  fdx.str(string("New entry------------------------------------------------>",13))
  i:=fat0.fileSeek(seek_position)                       'seek to DIR entry (byte) position
  tmp:=fat0.getCurrentPosition
  fat0.readData(@databuf, dirsize)
    fdx.dec(tmp)
    fdx.tx(" ")
    fdx.dec(fat0.getSector)
    fdx.tx(" ")
    fdx.hex(databuf[0],2)                             
    fdx.str(string("..."))
    repeat j from 12 to dirsize-1
      fdx.hex(databuf[j],2)                             
      fdx.tx(" ")
    repeat j from 0 to dirsize-1
      fdx.tx(ascii(databuf[j]))
    fdx.tx(13)
'-----------------------------------------------------------------
'NOW STORE THE EXTENTS
  ex := databuf[12]
  s1 := databuf[13]
  s2 := databuf[14]
  rc := databuf[15]
  fdx.str(string("ex="))
  fdx.dec(ex)
  fdx.str(string(" s1="))
  fdx.dec(s1)
  fdx.str(string(" s2="))
  fdx.dec(s2)
  fdx.str(string(" rc="))
  fdx.dec(rc)
  repeat i from 0 to 7                                  'get 7 allocation block nos
    j := i << 1                                         ' 2*i
    al[i] := (databuf[17+j]<<8 | databuf[16+j]) << blockshift
    fdx.str(string(" al["))
    fdx.dec(i)
    fdx.str(string("]="))
    fdx.dec(al[i])
  fdx.tx(13)

'  if ex > 0
'    StopOnError(string("Files >32KB not supported"))

'-----------------------------------------------------------------
'NOW SEEK TO READ EACH EXTENT & DISPLAY/OUTPUT
'calc correct rc value
  rc := ((ex & extentmask) * 128) + rc                  ' 128 = $80 (256 will be all extents)
  total_records += rc                                   'accumulate total 4KB records
  fdx.str(string("      Records= "))
  fdx.dec(rc)
  fdx.str(string(" ("))
  i := rc >> 3                                          ' /8 converts to KB
  fdx.dec(i)
  fdx.str(string("KB)"))
  crlf

  i := 0
  repeat while (rc > 0) & (i < 8)                       'read up to 8 @ 4KB blocks (256 sectors)
    if rc > 31
      readcpmblock(al[i] + dirbase, 32)                 'read up to a 4KB block
      rc := rc - 32                                     'prep for next loop
    else
      readcpmblock(al[i] + dirbase, rc)
      rc := 0                                           'all done
    i++


PRI readcpmblock(posn,recs) |  i, j, ctlz, tmp, lastrec
'' Read a block of data and display/write it

'NOW SEEK TO FIRST EXTENT
  i:=fat0.fileSeek(posn * sectorsize)
  lastrec := recs - 1
  repeat i from 0 to recs - 1
    tmp:=fat0.getCurrentPosition
    fat0.readData(@databuf, sectorsize)                 'get first data sector of block
{-
    fdx.dec(tmp)
    fdx.tx(" ")
    fdx.dec(fat0.getSector)
    fdx.tx(" ")
-}


'do not scan for Ctl-Z (therefore works for text and binaries)
    repeat j from 0 to 63                              
'     fdx.tx(ascii(databuf[j]))
'   fdx.str(string(13,"             "))
    repeat j from 64 to 127
'     fdx.tx(ascii(databuf[j]))
'   fdx.tx(13)
''  repeat j from 0 to 127
''    fat1.writeByte(databuf[j]
    fat1.writeData(@databuf, 128)

{---
'should scan backwards for <>Ctl-Z !!!!!!!!
    if i <> lastrec                                     'not last rec
      repeat j from 0 to 63                              
'       fdx.tx(ascii(databuf[j]))
        fat1.writeByte(databuf[j])
'     fdx.str(string(13,"             "))
      repeat j from 64 to 127
'       fdx.tx(ascii(databuf[j]))
        fat1.writeByte(databuf[j])
'     fdx.tx(13)
    else                                                'last rec
      ctlz := false                                     'preset not found
      repeat j from 0 to 63                             'stops when Ctl-Z found
        if databuf[j] == $1A
          ctlz := true
        if !ctlz
'         fdx.tx(ascii(databuf[j]))
          fat1.writeByte(databuf[j])
'     fdx.str(string(13,"eof          "))
      repeat j from 64 to 127
        if databuf[j] == $1A
          ctlz := true
        if !ctlz
'         fdx.tx(ascii(databuf[j]))
          fat1.writeByte(databuf[j])
'     fdx.tx(13)
---}
       
PRI ASCII(ch)
  result := ch & $7F                                    'remove top bit
  if result < $20
    result := "."

PRI StopOnError(msg)
  crlf
  fdx.str(msg)
  crlf
  fdx.str(string("Program Stop."))
  repeat                                                '<===== loop

PRI crlf
  fdx.tx(13)

{{
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |                                                            
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    | 
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
}} 